#include <dos.h>
#include <stdio.h>
#include <string.h>
#include <process.h>
#include <time.h>

#include <ctype.h>
#include <sys\types.h>
#include <sys\stat.h>
#include <malloc.h>
#include <direct.h>

#include "gen.h"
#include "sh.h"


/*
 *	sh:  a Unix(tm) like command interpreter for MS-DOS.
 *
 *  by  Douglas Orr
 *      Textset Inc.
 *      Ann Arbor, Michigan
 *
 *  Copyright (c) 1985 Textset.   All rights reserved.
 *
 *  This program may be freely distributed, but not sold for profit.
 *
 */


Public char * history[];
Public short  hindex[];
Public short hind, hmax;

Local char * mons[] =
	{
	"Inv", 
	"Jan", "Feb", "Mar",
	"Apr", "May", "Jun",
	"Jul", "Aug", "Sep",
	"Oct", "Nov", "Dec",
	};

char *
smalloc( size )
unsigned size;
	{
	char	* addr;

	if( !(addr = malloc(size)) )
		{
		/*  death with honor */
		fprintf( stderr, "out of space\n" );
		exit( 1 );
		}
	return( addr );

	}

char *
strdup( str )
Reg char * str;
	{
	return( strcpy( smalloc( strlen(str)+1 ), str ) );
	}




fcmp( file1, file2 )
char * * file1;
char * * file2;
	{
	return( strcmp( *file1, *file2 ) );
	}


/*		Built-in  commands		*/

/*	toy ls  */

struct  fs {
	char * fs_name;
	struct tm fs_time;
	unsigned short fs_mode;
	off_t fs_size;
	};


#define	Set(flag) (flags |= (1 << flag))
#define	Reset(flag) (flags &= ~(1 << flag))
#define Test(flag) (flags & (1 << flag))
#define	Recurse (0)
#define	Lflag (1)
#define	Sflag (2)
#define	Rflag (3)
#define	Aflag (4)
#define	Cflag (5)


Local
fscmp( file1, file2 )
struct fs * file1;
struct fs * file2;
	{
	return( strcmp( file1->fs_name, file2->fs_name ) );
	}

Local long ls_size;
Local int ls_cnt;

b_ls( argv, envp )
char * * argv;
char * * envp;
	{
	char * ptr;
	short flags = 0;
	int rc;

#ifdef	correct
	/*
	 * !! as near as I can tell, ls is the only built-in that allocates
	 *    storage.  If you interrupt, that storage doesn't get freed,
	 *    currently.  One way around this is the following, which isn't
	 *    very satisfactory.  Another is to just ignore it and hope you
	 *    don't run out of storage.  Yet another would be to set another
	 *    interrupt which frees the storage before leaving which is more
	 *    of a hassle than I have time for today.  Caveat user.
	 */
	signal( SIGINT, SIG_IGN );
#endif	correct


	ls_size = 0;
	ls_cnt = 0;

	Set(Recurse);		/* list subdirectories */

	if( !Interactive() )
		Set(Cflag);	/* print out one column */

	while( (*++argv)[0] == '-' )
		{
		for( ptr=(*argv)+1; *ptr; ptr++ )
		switch( *ptr )
			{
			case 'l':
				Set(Lflag);
				Reset(Cflag);
				break;
			case 's':
				Set(Sflag);
				break;
			case 'R':
				Set(Rflag);
				break;
			case 'a':
				Set(Aflag);
				break;
			default:
				fprintf( stderr, "ls: illegal option %s\n", 
					*argv );
				return( -1 );

			}
		}

	rc = ls( argv, flags );

	if( Test(Sflag) )
		fprintf( stdout, "%ld total.  %d file%s.\n", ls_size, ls_cnt,
		ls_cnt == 1 ? "" : "s" );

	return( rc );
	}



/*
 *   UGLY ALERT  ... since I couldn't find any good way to do what I
 *   would do on Unix, were I to be building a program to recursively
 *   search directories - that being open the damn things like files,
 *   I debased myself (a frequent occurance in dosville) and issue "cd"s
 *   to move around.  My apologies.  I do know better, in the global
 *   sense.
 */

Local
lstatus( file, flags, maxlen, nl )
struct	fs * file;
short flags;
int maxlen;
bool nl;
	{

	/*   UGLY ALERT   recurse on directories if appropriate  */
	/*  (Jaayynnnee ... how do you stop this crazy thing?) */
	if( (file->fs_mode & S_IFDIR) && Test(Recurse) )
		{
		int dev = -1;
		char * auxlist[1];
		char * cwd = NULL;
		char * newd;
		char * dir = file->fs_name;

		/*   cheat   */
		if( !Test(Rflag) )
			Reset(Recurse);	/* only recurse one level, by default  */

		auxlist[0] = NULL;

		if( dir[1] == ':' )
			{
			dev = getcdsk();
			if( chdsk( dir[0] ) == -1)
				{
				fprintf( stderr, "couldn't access %c:\n",  dir[0] );
				return( -1 );
				}
			dir += 2;
			if( dir[0] == '\0' )
				dir = "/";
			}

		cwd = getcwd(NULL,128);
		if( (cwd == NULL) || chdir(dir) == -1)
			{
			fprintf( stderr, "! couldn't cd to %s\n",  file->fs_name );
			return( -1 );
			}

		if( (newd = getcwd(NULL,128)) != NULL )
			{
			fprintf( stdout, "%s\n%s:\n", (nl) ? "" : "\n", strlwr(newd) );
			free(newd);
			}

		/*  do a list of the place to which we just attached  */
		ls( auxlist, flags );
		putc( '\n', stdout );

		if( dev != -1 )
			chdsk( dev );
		if( cwd )
			{
			chdir( cwd ); free( cwd );
			}

		nl = True;

		}
	else
		{
		if( Test(Lflag) )
			{
			/* ?! doesn't always line up right  */
			fprintf( stdout, "%-8ld  %2d:%02d   %3s-%02d-%02d    %-13s\n", 
				file->fs_size,  file->fs_time.tm_hour,
			 	file->fs_time.tm_min, 
				mons[file->fs_time.tm_mon], 
				file->fs_time.tm_mday,
				file->fs_time.tm_year, file->fs_name );
			nl = True;
			}
		else
			{
			fprintf( stdout, "%s", file->fs_name );

			if( Test(Aflag) )
				{
				if( file->fs_mode & S_IFDIR )
					fprintf( stdout, "\\" );
				else
				if( file->fs_mode & S_IEXEC )
					fprintf( stdout, "*" );
				else
					printf( " " );
				}

			fprintf( stdout, "%*s", maxlen-strlen(file->fs_name), 
				"" );
			nl = False;
			}
		}

	return( nl );

	}

Local
ls( list, flags )
char * * list;
short flags;
	{

	int argc;
	struct  fs * files;
	char * * filelist;

	int cnt;
	int maxlen;
	Reg int i, j;
	Reg int r;
	int cols;
	char * ptr;

	Local struct stat statb;
	int nl;

	/*  initialize  */

	for( argc=0, filelist=list ; *filelist; filelist++ )
		argc++;

	maxlen = 0;
	files = NULL;
	cnt = 0;

	/*   implicitly ls of current directory  */
	if( argc == 0 )
		{
		int ac;
		files = (struct fs *)smalloc( (ac=20) * sizeof(struct fs) );

		if( !Test(Rflag) )
			Reset(Recurse);

		for( ptr=open_dir("*.*"); ptr; ptr=nxt_entry() )
			{

			if( cnt == ac )
				{
				ac += 20;
				files = (struct fs *)realloc( files, ac*sizeof(struct fs) );
				if( files == NULL )
					{
					/*  cut your losses */
					fprintf( stderr, "ls: out of space\n" );
					return(-1);
					}
				}

			/* degenerate */
			if( strcmp( ptr, "." ) == 0 || strcmp( ptr, ".." ) == 0 )
				continue;

			files[cnt].fs_mode = dta_mode();
			files[cnt].fs_time = *dta_time();
			files[cnt].fs_size = dta_size();
			files[cnt++].fs_name = strlwr(strdup(ptr));

			if( strlen(ptr) > maxlen )
					maxlen = strlen( ptr );

			}

		}
	else
		{

		files = (struct fs *)smalloc( (argc+1) * sizeof(struct fs) );
		filelist = list;

		while( *filelist )
			{

			if( (stat( *filelist, &statb ) == -1) )
				{
				if( isdev(*filelist)
				|| (strcmp(*filelist, "/")==0 
				||  strcmp(*filelist,"\\")==0))
					{
					statb.st_mode = S_IFDIR;
					statb.st_size = 0;
					statb.st_mtime = 0;
					}
				else
					{
					fprintf( stderr, "%s: no such file\n", 
						*filelist );
					return( -1 );
					}
				}

			files[cnt].fs_name = strlwr(strdup(*filelist));
			files[cnt].fs_mode = statb.st_mode;
			files[cnt].fs_time = *localtime( &statb.st_mtime );
			files[cnt].fs_time.tm_mon += 1;    /* relative to 1 */
			files[cnt++].fs_size = statb.st_size;

			if( strlen(*filelist) > maxlen )
					maxlen = strlen( *filelist );

			filelist++;	  /* very important  */

			}
		}

	/*  !!  this can overflow stack if the list is too large  */
	if( cnt )
		qsort( (char *)files, cnt, sizeof(struct fs), fscmp );


	/*  formatting considerations  */
	maxlen++;
	if( Test(Aflag) )
		cols = 79 / (maxlen+1);
	else
		cols = 79 / maxlen;

	if( cols < 0 )
		cols = 1;

	if( Test(Lflag) )
		{
		for( i=0; i<cnt; i++ )
			(void)lstatus( &files[i], flags, maxlen, True );
		}
	else
	if( Test(Cflag) )
		{
		for( i=0; i<cnt; i++ )
			fprintf( stdout, "%s\n", files[i].fs_name );
		}
	else
		{
		int ind;

		/* print out by rows */
		nl = True;
		for( r=0; r < ((cnt+cols-1)/cols); r++ )
			{

			for( ind=r, i=0; i<cols; i++ )
				{

				/*  have we seen everything? */
				if( ((r*cols) + i) >= cnt )
					break;

				nl = lstatus( &files[ind], flags, maxlen, nl );

				/* add the size of a column */
				ind += cnt/cols;  
				if( i < (cnt % cols) ) 
					ind++;

				}

			/*  nl flag indicates if a new line is desired */
			if( !nl )
				{
				putc( '\n', stdout );
				nl = True;
				}

			}
		}

	/*  free up the storage you have used, run some totals */
	ls_cnt += cnt;

	for( i=0; i<cnt; i++ )
		{
		ls_size += files[i].fs_size;
		free( files[i].fs_name );
		}

	if( files )
		free( files );

	return( 0 );

	}


b_echo( argv, envp )
char * * argv;
char * * envp;
	{

	int	cnt = 0;

	argv++;
	while( *argv )
		{
		if( cnt++ != 0 )
			fprintf( stdout, " " );
		fprintf( stdout, "%s", *argv );
		argv++;
		}

	putc( '\n', stdout );
	return( 0 );

	}

b_cd( argv, envp )
char * * argv;
char * * envp;
	{

	argv++;

	if( strlen(*argv) == 2
	&&  (*argv)[1] == ':' )
		return( chdsk( *argv ) );

	if( chdir( *argv ) == -1 )
		perror( "chdir" );
	return( 0 );

	}

b_pwd( argv, envp )
char * * argv;
char * * envp;
	{

	char * dirp;

	if( (dirp = getcwd(NULL,128)) == NULL )
		{
		perror( "getcwd" );
		return( -1 );
		}
	else
		{
		fprintf( stdout, "%s\n", strlwr(dirp) );
		free( dirp );
		return( 0 );
		}

	}


#define	MaxDirs (10)
Local char * dirs[MaxDirs];
Local int dircnt = 0;

b_dirs( argv, envp )
char * * argv;
char * * envp;
	{

	Reg int i;
	char * cwd;

	if( (cwd = getcwd( NULL, 128 )) )
		{
		fprintf( stdout, "%s ", strlwr(cwd) );
		free( cwd );
		}

	for( i=dircnt-1; i >= 0; i-- )
		fprintf( stdout, "%s ", dirs[i] );
	putc( '\n', stdout );

	return( 0 );

	}

b_pushd( argv, envp )
char * * argv;
char * * envp;
	{

	char * dirp;
	char * cwd;

	if( *++argv == NULL )
		{
		if( dircnt == 0 )
			{
			fprintf( stderr, "no directory specified\n" );
			return( -1 );
			}
		else
			dirp = dirs[--dircnt];
		}
	else
		dirp = *argv;

	if( dircnt < MaxDirs )
		{

		if( (cwd = getcwd( NULL, 128 )) == NULL )
			{
			perror( "chdir" );
			return( -1 );
			}

		if( dirp[1] == ':' )
			{
			if( chdsk(dirp[0]) == -1 )
				perror( "chdisk" );
			dirp += 2;
			if( dirp[0] == '\0' )
				dirp = "/";
			}

		if ( chdir( dirp ) == -1 )
			{
			perror( "chdir" );
			free(cwd);
			}
		else
			dirs[dircnt++] = strlwr(cwd);

		b_dirs( NULL, NULL );
		return( 0 );

		}
	else
		fprintf( stderr, "stack overflow\n" );

	return( -1 );

	}


b_popd( argv, envp )
char * * argv;
char * * envp;
	{

	char * cwd;
	char * dir;

	if( dircnt == 0 )
		{
		fprintf( stderr, "stack empty\n" );
		return( -1 );
		}
	else
		{
		dir = cwd = dirs[--dircnt];
		if( dir[1] == ':' )
			{
			if( chdsk(dir[0]) == -1 )
				perror( "chdisk" );
			dir += 2;
			}

		if( chdir( dir )  == -1)
			perror( "chdir:" );
		else
			b_dirs( NULL, NULL );

		free( cwd );
		}
	return( 0 );

	}

b_mkdir( argv, envp )
char * * argv;
char * * envp;
	{

	int error = 0;
	int rc;

	while( *++argv )
		{
		if( (rc = mkdir( *argv )) != 0 )
			perror( "mkdir" );
		error |= rc;
		}
	return( error );

	}


b_rmdir( argv, envp )
char * * argv;
char * * envp;
	{

	int error = 0;
	int rc;

	while( *++argv )
		{
		if( (rc = rmdir( *argv )) != 0 )
			perror( "rmdir" );
		error |= rc;
		}
	return( error );

	}


b_rm( argv, envp )
char * * argv;
char * * envp;
	{

	int	error = 0;
	int rc;

	while( *++argv )
		{
		if( (rc = unlink( *argv )) != 0 )
			perror( "rm" );
		error |= rc;
		}
	return( error );

	}

b_history( argv, envp )
char * * argv;
char * * envp;
	{

	int count;
	int i;
	int ind;

	if( *++argv == NULL )
		count = hmax;
	else
		{
		count = atoi( *argv );
		if( count > hmax ) count = hmax;
		}

	ind = hind-1;
	for( i=0; i<count; i++ )
		{
		if( ind < 0 )
			ind = hmax-1;

		if( history[ind] == NULL )
			break;

		fprintf( stdout, "%d: %s\n", hindex[ind], history[ind] );
		ind--;
		}

	return( 0 );

	}

/* ahhbedee ahhbedee ahhba  That's All Folks  */
b_exit(argv, envp)
char * * argv;
char * * envp;
	{
	if( *++argv )
		exit( atoi(*argv) );
	else
		exit(0);
	}

b_rehash( argv, envp )
char * * argv;
char * * envp;
	{

	rehash( l_getenv( "PATH" ) );
	return( 0 );

	}

b_set( argv, envp )
char * * argv;
char * * envp;
	{
	char * vptr;

	if( *++argv == NULL )
		l_printenv( envp );
	else
		{
		while( *argv )
			{
			if( (vptr = strchr( *argv, '=' )) == NULL )
				{
				fprintf( stderr, "set: syntax error\n" );
				return( -1 );
				}

			*vptr++ = '\0';
			if( l_setenv( *argv, vptr ) == -1 )
				{
				perror( "set" );
				return( -1 );
				}
			argv++;
			}
		return( 0 );
		}

	}


b_source( argv, envp )
char * * argv;
char * * envp;
	{
	while( *++argv )
		source( *argv );
	}


/*
 *   If you're wondering what fgrep is doing here, it's because the
 *  goddamn msdos arg list is so small.  Anything that I do frequently
 *  is a builtin command, since you can't adequately wildcard expand
 *  on large file lists.  Every day, in every way, msdos sucks.
 */

Local
fgrep( file, pattern, fd )
char	* file;
char	* pattern;
FILE *	fd;
	{

	char line[BUFSIZ];
	int rc;

	rc = 0;
	while( fgets( line, sizeof(line), fd ) != NULL )
		if( rc = match( pattern, line ) )
			fprintf( stdout, "%s: %s", 
				(file) ? file : "stdin", line );

	return( rc );

	}

Local
match( pattern, line )
char	* pattern;
char	* line;
	{

	/*  !? not a great algorithm  */
	char	* ptr;
	char	* end;
	int	plen = strlen(pattern);
	int llen = strlen(line);

	if( plen > llen )
		return( 0 );

	end = line+(llen-plen);

	for( ptr=line; ptr < end; ptr++ )
		{
		/*   ... actually, pretty mediocre  */
		if( strncmp( pattern, ptr, plen ) == 0 )
			return( 1 );
		}

	return( 0 );

	}

b_fgrep( argv, envp )
char * argv[];
char * * envp;
	{

	FILE *  fd;
	int rc;
	char * pattern;

	pattern = *++argv;
	argv++;

	rc = 0;
	if( *argv == NULL )
		{
		rc = fgrep( NULL, pattern, stdin );
		return( 0 );
		}
	else
	while( *argv )
		{
		if( (fd = fopen( *argv, "r" )) == NULL )
			fprintf( stderr, "couldn't open %s\n", *argv );
		else
			{
			rc |= fgrep( *argv, pattern, fd );
			fclose( fd );
			}
		argv++;
		}

	return( !rc );

	}
